"""CLAM: Clustered Learning of Approximate Manifolds."""

from . import anomaly_detection
from . import classification
from . import core
from . import search
from . import utils
from .core import Cluster
from .core import ClusterCriterion
from .core import Dataset
from .core import Edge
from .core import Graph
from .core import GraphCriterion
from .core import Metric
from .core import Space
from .core import cluster
from .core import cluster_criteria
from .core import dataset
from .core import graph
from .core import graph_criteria
from .core import metric
from .core import space

__version__ = "0.14.0"



"""CHAODA: Clustered Hierarchical Anomaly and Outlier Detection Algorithms."""

import inspect
import math
import typing

import numpy

from .. import core
from .. import search
from ..core import cluster_criteria
from ..core import graph_criteria
from ..utils import constants
from ..utils import helpers
from . import graph_scorers
from . import pretrained_models

logger = helpers.make_logger(__name__)

VotingMode = typing.Literal[
    "mean",
    "product",
    "median",
    "min",
    "max",
    "p25",
    "p75",
]

DEFAULT_GRAPH_CRITERIA = [
    graph_criteria.MetaMLSelect(function, min_depth=4)
    for _, function in inspect.getmembers(pretrained_models, inspect.isfunction)
]

DEFAULT_GRAPH_SCORERS: list[graph_scorers.GraphScorer] = [
    graph_scorers.ClusterCardinality(),
    graph_scorers.ComponentCardinality(),
    graph_scorers.VertexDegree(),
    graph_scorers.ParentCardinality(depth_weight=lambda d: 1 / (d**0.5)),
    graph_scorers.GraphNeighborhood(eccentricity_fraction=0.25),
    graph_scorers.StationaryProbabilities(steps=16),
]


class CHAODA:
    """This class is the main CHAODA class."""

    def __init__(  # noqa: PLR0913
        self,
        metric_spaces: typing.Sequence[core.Space],
        *,
        partition_criteria: typing.Optional[
            typing.Sequence[cluster_criteria.ClusterCriterion]
        ] = None,
        selector_scorers: typing.Optional[
            typing.Sequence[
                tuple[
                    graph_criteria.GraphCriterion,
                    typing.Sequence[graph_scorers.GraphScorer],
                ]
            ]
        ] = None,
        normalization_mode: helpers.NormalizationMode = "gaussian",
        use_speed_threshold: bool = True,
        voting_mode: VotingMode = "mean",
    ) -> None:
        """Creates and initializes a CHAODA object.

        Args:
            metric_spaces: A list of metric spaces to use for anomaly detection.
                All metric spaces should have the same `Dataset` object.
            partition_criteria: list of criteria for partitioning clusters when
                building trees.
            selector_scorers: list of 2-tuples whose items are
                - a trained meta-ml model for selecting a graph.
                - a list of individual algorithms to run on that graph.
            normalization_mode: What normalization mode to use. Must be one of
                - 'linear',
                - 'gaussian', or
                - 'sigmoid'.
            use_speed_threshold: Whether to skip slow graph scorers.
            voting_mode: to use to aggregate scores for the ensemble.
        """
        for i, left in enumerate(metric_spaces):
            for right in metric_spaces[i + 1 :]:
                if left.data != right.data:
                    msg = (
                        f"Metric spaces {left.name} and {right.name} "
                        "had different datasets."
                    )
                    raise ValueError(
                        msg,
                    )

        self.__metric_spaces = metric_spaces

        self.__partition_criteria: typing.Sequence[
            cluster_criteria.ClusterCriterion
        ] = partition_criteria or [
            cluster_criteria.MinPoints(
                int(math.log2(self.__metric_spaces[0].data.cardinality)),
            ),
        ]

        self.__selector_scorers: typing.Sequence[
            tuple[
                graph_criteria.GraphCriterion,
                typing.Sequence[graph_scorers.GraphScorer],
            ]
        ] = selector_scorers or [
            (c, [s for s in DEFAULT_GRAPH_SCORERS if s.name in c.name])
            for c in DEFAULT_GRAPH_CRITERIA
        ]

        self.__normalization_mode = normalization_mode
        self.__use_speed_threshold = use_speed_threshold
        self.__voting_mode = voting_mode

        self.__bowls = [
            SingleSpaceChaoda(
                metric_space,
                self.__partition_criteria,
                self.__selector_scorers,
                self.__normalization_mode,
                self.__use_speed_threshold,
            )
            for metric_space in self.__metric_spaces
        ]

        self.__scores: typing.Union[numpy.ndarray, constants.Unset] = constants.UNSET

    def build(self) -> "CHAODA":
        """Build the CHAODA model and compute anomaly scores."""
        self.__bowls = [bowl.build() for bowl in self.__bowls]
        individual_scores = numpy.concatenate([bowl.scores for bowl in self.__bowls])
        self.__scores = self.__vote(individual_scores)
        return self

    def __vote(self, scores: numpy.ndarray) -> numpy.ndarray:
        """Vote among individual scores for ensemble score."""
        if self.__voting_mode == "mean":
            scores = numpy.mean(scores, axis=0)
        elif self.__voting_mode == "product":
            scores = numpy.product(scores, axis=0)
        elif self.__voting_mode == "median":
            scores = numpy.median(scores, axis=0)
        elif self.__voting_mode == "min":
            scores = numpy.min(scores, axis=0)
        elif self.__voting_mode == "max":
            scores = numpy.max(scores, axis=0)
        elif self.__voting_mode == "p25":
            scores = numpy.percentile(scores, 25, axis=0)
        else:  # self.__voting_mode == 'p75'
            scores = numpy.percentile(scores, 75, axis=0)

        return scores

    @property
    def scores(self) -> numpy.ndarray:
        """The ensemble scores for each point in the dataset."""
        if self.__scores is constants.UNSET:
            msg = "Please call the `fit` method before using this property."
            raise ValueError(msg)
        return self.__scores

    def fit_predict(self) -> numpy.ndarray:
        """Fit the model and return the anomaly scores."""
        self.build()
        return self.__scores

    def predict_single(self, query: typing.Any) -> float:  # noqa: ANN401
        """Predict the anomaly score for a single query."""
        scores = []
        for bowl in self.__bowls:
            for cluster_scores in bowl.cluster_scores_list:
                hits = set(bowl.searcher.tree_search_history(query, 0.0)[0].keys())

                intersection = hits.intersection(set(cluster_scores.keys()))
                if len(intersection) > 0:
                    score = self.__vote(
                        numpy.asarray(
                            [cluster_scores[c] for c in intersection],
                            dtype=numpy.float32,
                        ),
                    )
                else:
                    score = 1.0
                scores.append(1.0 if numpy.isnan(score) else float(score))

        final_score = self.__vote(numpy.asarray(scores, dtype=numpy.float32))
        return float(final_score)

    def predict(self, queries: core.Dataset) -> numpy.ndarray:
        """Predict the anomaly scores for a 2d array of queries."""
        scores = []
        for i in range(queries.cardinality):
            logger.info(f"Predicting anomaly score for query {i} ...")
            scores.append(self.predict_single(queries[i]))
        return numpy.asarray(scores, dtype=numpy.float32)


class SingleSpaceChaoda:
    """This class is a single-metric-space version of CHAODA.

    This is intended to be used from the CHAODA class and does almost no input
    validation.

    This class does not vote among scores to form an ensemble.
    """

    def __init__(  # noqa: PLR0913
        self,
        metric_space: core.Space,
        partition_criteria: typing.Sequence[cluster_criteria.ClusterCriterion],
        selectors_scorers: typing.Sequence[
            tuple[
                graph_criteria.GraphCriterion,
                typing.Sequence[graph_scorers.GraphScorer],
            ]
        ],
        normalization_mode: helpers.NormalizationMode,
        use_speed_threshold: bool,
    ) -> None:
        """Creates single-metric-space version of chaoda.

        Args:
            metric_space: The metric space to use.
            partition_criteria: A list of criteria to use for partitioning
                clusters.
            selectors_scorers: list of 2-tuples whose items are
                - a trained meta-ml model for selecting clusters to build
                  graphs.
                - a list of individual algorithms for predicting anomaly scores
                  from that graph.
            normalization_mode: method to use to normalize anomaly scores from
                different scoring algorithms.
            use_speed_threshold: Whether to only run the fast algorithms in the
                ensemble.
        """
        self.__metric_space = metric_space
        self.__partition_criteria = partition_criteria
        self.__selectors = [s for s, _ in selectors_scorers]
        self.__scorers = [s for _, s in selectors_scorers]
        self.__normalization_mode = normalization_mode
        self.__use_speed_threshold = use_speed_threshold

        self.__root: typing.Union[core.Cluster, constants.Unset] = constants.UNSET
        self.__graphs: typing.Union[list[core.Graph], constants.Unset] = constants.UNSET
        self.__cluster_scores_list: typing.Union[
            list[graph_scorers.ClusterScores],
            constants.Unset,
        ] = constants.UNSET
        self.__scores: typing.Union[numpy.ndarray, constants.Unset] = constants.UNSET
        self.__searcher: typing.Union[search.CAKES, constants.Unset] = constants.UNSET

    @property
    def root(self) -> core.Cluster:
        """The root cluster of the tree."""
        if self.__root is constants.UNSET:
            msg = "Please call the `fit` method before using this property."
            raise ValueError(msg)
        return self.__root  # type: ignore[return-value]

    @property
    def scores(self) -> numpy.ndarray:
        """The individual scores for each point in the dataset."""
        if self.__scores is constants.UNSET:
            msg = "Please call the `fit` method before using this property."
            raise ValueError(msg)
        return self.__scores  # type: ignore[return-value]

    @property
    def cluster_scores_list(self) -> list[graph_scorers.ClusterScores]:
        """The list of cluster scores for each graph."""
        if self.__cluster_scores_list is constants.UNSET:
            msg = "Please call the `fit` method before using this property."
            raise ValueError(msg)
        return self.__cluster_scores_list  # type: ignore[return-value]

    @property
    def searcher(self) -> search.CAKES:
        """The CAKES object for this metric space."""
        if self.__searcher is constants.UNSET:
            msg = "Please call the `fit` method before using this property."
            raise ValueError(msg)
        return self.__searcher  # type: ignore[return-value]

    def build(self) -> "SingleSpaceChaoda":
        """Build a Manifold on the data, create optimal Graphs, and predict scores."""
        self.__root = (
            core.Cluster.new_root(self.__metric_space)
            .build()
            .iterative_partition(self.__partition_criteria)
            .normalize_ratios(mode=self.__normalization_mode)
        )
        self.__searcher = search.CAKES.from_root(self.__root)

        self.__graphs = [
            core.Graph(selector(self.__root)).build() for selector in self.__selectors
        ]

        if self.__use_speed_threshold:
            individual_scores = [
                method(g)
                for g, scorers in zip(self.__graphs, self.__scorers)
                for method in scorers
                if method.should_be_fast(g)
            ]
            if len(individual_scores) == 0:
                uniform_cluster_scores = {self.__root: 0.5}
                uniform_scores = 0.5 * numpy.ones(shape=(self.__root.cardinality,))
                individual_scores.append((uniform_cluster_scores, uniform_scores))
        else:
            individual_scores = [
                method(g)
                for g, scorers in zip(self.__graphs, self.__scorers)
                for method in scorers
            ]

        self.__scores = numpy.stack([s for _, s in individual_scores])
        self.__cluster_scores_list = [d for d, _ in individual_scores]

        return self


__all__ = [
    "VotingMode",
    "CHAODA",
    "SingleSpaceChaoda",
]



"""Utilities for training CHAODA."""

import pathlib
import typing

import numpy

from .. import core
from ..core import graph_criteria
from ..utils import helpers
from . import anomaly_space
from . import graph_scorers
from . import meta_ml

logger = helpers.make_logger(__name__)

# metric -> scorer -> list of meta-ml models
ModelsDict = dict[str, dict[graph_scorers.GraphScorer, list[meta_ml.MetaMLModel]]]

# metric -> scorer -> data for meta-ml models
ScorerData = dict[str, dict[graph_scorers.GraphScorer, numpy.ndarray]]


def data_from_graph(
    cluster_scores: graph_scorers.ClusterScores,
    labels: numpy.ndarray,
) -> tuple[numpy.ndarray, numpy.ndarray]:
    """Extracts data from scored clusters to use for training meta-ml models.

    Args:
        cluster_scores: The dict of cluster scores obtained by running a scorer
        on a graph.
        labels: Ground-truth anomaly scores for the points in the graph.

    Returns:
        A 2-tuple whose elements are:
            - a 2d array of feature vectors, i.e. the cluster ratios.
            - a 1d array of (1 - loss) for each cluster.
    """
    clusters = list(cluster_scores.keys())
    scores = list(cluster_scores.values())

    train_x = numpy.zeros(shape=(len(clusters), 6), dtype=numpy.float32)
    train_y = numpy.zeros(shape=(len(clusters),))

    for i, (cluster, score) in enumerate(zip(clusters, scores)):
        train_x[i] = cluster.ratios

        y_true = numpy.asarray(labels[cluster.indices], dtype=numpy.float32)

        loss = (
            float(numpy.sqrt(numpy.mean(numpy.square(score - y_true))))
            / cluster.cardinality
        )

        train_y[i] = 1.0 - loss

    return train_x, train_y


def save_models(path: pathlib.Path, meta_models: ModelsDict) -> None:
    """Saves the models in the given file."""
    # list of tuples whose elements are:
    # - list of import lines
    # - code for function
    model_codes: list[tuple[list[str], str]] = [
        model.extract_python(metric_name, scorer.name)
        for metric_name, scorer_models in meta_models.items()
        for scorer, models in scorer_models.items()
        for model in models
    ]

    [import_lines_list, function_codes] = list(zip(*model_codes))
    import_lines: list[str] = list(
        {lines for import_lines in import_lines_list for lines in import_lines},
    )

    with path.open("w") as writer:
        writer.writelines(import_lines)
        writer.write("\n")

        for code in function_codes:
            writer.write(f"\n\n{code}\n")


def train_meta_ml(  # noqa: PLR0913
    *,
    spaces_criteria: typing.Sequence[
        tuple[anomaly_space.AnomalySpace, typing.Sequence[core.ClusterCriterion]]
    ],
    models_kwargs: typing.Sequence[
        tuple[type[meta_ml.MetaMLModel], dict[str, typing.Any]]
    ],
    scorers: typing.Sequence[graph_scorers.GraphScorer],
    out_dir: pathlib.Path,
    num_epochs: int = 10,
    save_frequency: int = 1,
    only_train_fast_scorers: bool = False,
) -> pathlib.Path:
    """Trains meta-ml models for CHAODA. See examples/chaoda_training.py for usage.

    Training for CHAODA takes the following steps:

    1. We build a cluster tree for each given metric space using the respective
    partition criteria. These trees are only built once. They are used to select
    graphs during training.

    2. We start the first training epoch. In this epoch, since we do not have
    pretrained meta-ml models for selecting graphs, we will select every fifth
    layer from the trees to use as graphs for training the models. For each
    subsequent epoch, we will use the meta-ml models from the previous epoch to
    select new graphs.

    3. During an epoch, from each tree, we use each graph selection criterion,
    i.e. fixed-depth layers for the first epoch and meta-ml models for other
    epochs, to select a graph.

    4. For each individual algorithm for graph scoring, we score the clusters in
    the graph. For each cluster, we compute the difference, i.e. the loss
    between the mean ground-truth scores of the points in that cluster and the
    predicted anomaly score from the graph. The training data for the models are
    six ratios for each cluster as the feature vector and (1 - loss) as the
    target.

    5. For each individual algorithm, we merge the training data from the
    current epoch with those data from previous epochs. We use these data to
    train the meta-ml models. Using data from previous epochs in this way
    ensures that models learn what cluster ratios make for good performance and
    what ratios make for poor performance on the anomaly detection task.

    6. After finishing training, and during training at the user specified
    frequency, we extract the decision functions from the trained meta-ml models
    and write them out as python functions. These exported functions can be used
    in `graph_criteria.MetaMLSelect` to select graphs for inference on new
    datasets.

    Args:
        spaces_criteria: list of 2-tuples whose items are:
            - an AnomalySpace
            - a list of partition criteria to use for building the root cluster
              for that space.
        models_kwargs: A list of 2-tuples whose items are:
            - The meta-ml model class
            - A dictionary of key-word arguments to use for initializing an
              object of that that class.
        scorers: A list of unique graph scorers (the individual algorithms).
        out_dir: Where the training meta-ml models will be scores.
        num_epochs: The number of epochs to train the models.
        save_frequency: Save the trained models after each of this many epochs.
        only_train_fast_scorers: Whether to use the `should_be_fast` method on
            `GraphScorer`s to avoid generating training data from slow
            algorithms.

    Returns:
        Path to file where the final models were saved.
    """
    roots = []
    for space, criteria in spaces_criteria:
        logger.info(f"Building root cluster for {space.name} ...")

        root = (
            core.Cluster.new_root(space)
            .build()
            .iterative_partition(criteria)
            .normalize_ratios("gaussian")
        )
        roots.append(root)

    metric_names = list({s.distance_metric.name for s, _ in spaces_criteria})

    meta_models: ModelsDict = {
        metric_name: {
            scorer: [model(**kwargs) for model, kwargs in models_kwargs]
            for scorer in scorers
        }
        for metric_name in metric_names
    }

    full_train_x: ScorerData = {
        metric_name: {scorer: None for scorer in scorers}
        for metric_name in metric_names
    }
    full_train_y: ScorerData = {
        metric_name: {scorer: None for scorer in scorers}
        for metric_name in metric_names
    }

    for epoch in range(1, num_epochs + 1):
        logger.info(f"Starting Epoch {epoch}/{num_epochs} ...")

        for root, (space, _) in zip(roots, spaces_criteria):
            root_name = root.metric_space.name
            logger.info(f"Epoch {epoch}/{num_epochs}: Using root {root_name} ...")

            metric_name = space.distance_metric.name
            labels = space.data.labels

            graphs: dict[graph_scorers.GraphScorer, list[core.Graph]]
            if epoch == 1:
                graphs_list = [
                    core.Graph(graph_criteria.Layer(d)(root)).build()
                    for d in range(5, root.max_leaf_depth, 5)
                ]
                graphs = {scorer: graphs_list for scorer in scorers}
            else:
                graphs = {
                    scorer: [
                        core.Graph(
                            graph_criteria.MetaMLSelect(
                                lambda ratios: model.predict(  # noqa: B023
                                    ratios[None, :],
                                ),
                                name=model.name,
                            )(root),
                        ).build()
                        for model in models
                    ]
                    for scorer, models in meta_models[metric_name].items()
                }

            for scorer in scorers:
                logger.info(
                    f"Epoch {epoch}/{num_epochs}: Using root "
                    f"{root_name} and scorer {scorer.name} ...",
                )

                for graph in graphs[scorer]:
                    if only_train_fast_scorers and (not scorer.should_be_fast(graph)):
                        continue

                    (cluster_scores, _) = scorer(graph)
                    train_x, train_y = data_from_graph(cluster_scores, labels)

                    if epoch == 1:
                        full_train_x[metric_name][scorer] = train_x
                        full_train_y[metric_name][scorer] = train_y
                    else:
                        full_train_x[metric_name][scorer] = numpy.concatenate(
                            [full_train_x[metric_name][scorer], train_x],
                            axis=0,
                        )
                        full_train_y[metric_name][scorer] = numpy.concatenate(
                            [full_train_y[metric_name][scorer], train_y],
                            axis=0,
                        )

                new_models = []
                for model in meta_models[metric_name][scorer]:
                    logger.info(
                        f"Epoch {epoch}/{num_epochs}: Fitting model "
                        f"{model.name}, scorer {scorer.name} with root "
                        f"{root_name} and scorer {scorer.name} ...",
                    )
                    new_models.append(
                        model.fit(
                            full_train_x[metric_name][scorer],
                            full_train_y[metric_name][scorer],
                        ),
                    )
                meta_models[metric_name][scorer] = new_models

        if epoch % save_frequency == 0:
            logger.info(f"Saving models after epoch {epoch}/{num_epochs} ...")
            save_models(out_dir.joinpath(f"models_epoch_{epoch}.py"), meta_models)

    final_path = out_dir.joinpath("models_final.py")
    save_models(final_path, meta_models)

    return final_path


__all__ = [
    "train_meta_ml",
]



"""CHAODA: Clustered Hierarchical Anomaly and Outlier Detection Algorithms."""

from . import anomaly_dataset
from . import anomaly_space
from . import chaoda
from . import graph_scorers
from . import meta_ml
from . import pretrained_models
from . import training
from .anomaly_dataset import AnomalyDataset
from .anomaly_space import AnomalySpace
from .chaoda import CHAODA



"""Defines criteria for when to partition clusters."""

import abc

from ..utils import helpers
from . import cluster

logger = helpers.make_logger(__name__)


class ClusterCriterion(abc.ABC):
    """A rule to decide when a cluster can be partitioned.

    Subclasses must implement the `__call__` method to take a single `Cluster`
    and return a `bool` to indicate whether that cluster can be partitioned.

    If multiple criteria are used during `partition`, they must all return
    `True` to allow the cluster to be partitioned.
    """

    @abc.abstractmethod
    def __call__(self, c: "cluster.Cluster") -> bool:
        """Returns `True` if the cluster can be partitioned."""
        pass


class MaxDepth(ClusterCriterion):
    """Clusters with `depth` less than `max_depth` may be partitioned."""

    def __init__(self, max_depth: int) -> None:
        """Initializes the criterion with the given `max_depth`."""
        self.max_depth = max_depth

    def __call__(self, c: "cluster.Cluster") -> bool:
        """Returns `True` if the cluster is shallower than `max_depth`."""
        return c.depth < self.max_depth


class MinPoints(ClusterCriterion):
    """Clusters with `cardinality` greater than `min_points` may be partitioned."""

    def __init__(self, min_points: int) -> None:
        """Initializes the criterion with the given `min_points`."""
        self.min_points = min_points

    def __call__(self, c: "cluster.Cluster") -> bool:
        """Returns `True` if the cluster has more points than `min_points`."""
        return c.cardinality > self.min_points


class NotSingleton(ClusterCriterion):
    """Clusters can be partitioned so long as they are not singletons."""

    def __call__(self, c: "cluster.Cluster") -> bool:
        """Returns `True` if the cluster is not a singleton."""
        return not c.is_singleton


__all__ = [
    "ClusterCriterion",
    "MaxDepth",
    "MinPoints",
]



"""Constants used in the CLAM package."""

import logging
import os

LOG_LEVEL = getattr(logging, os.environ.get("CLAM_LOG", "INFO"))
EMA_SMOOTHING = float(os.environ.get("CLAM_EMA_SMOOTHING", "2"))
EMA_PERIOD = float(os.environ.get("CLAM_EMA_PERIOD", "10"))
EMA_ALPHA = EMA_SMOOTHING / (1 + EMA_PERIOD)

SUBSAMPLE_LIMIT = 100
BATCH_SIZE = 10_000
EPSILON = 1e-6


class Unset:
    """This is a hack around type-hinting when a value cannot be set in __init__.

    https://peps.python.org/pep-0661/

    Usage:

    ```python

    class MyClass:
        def __init__(self, *args, **kwargs):
            ...
            self.__value: typing.Union[ValueType, Unset] = UNSET

        def value_setter(self, *args, **kwargs):
            ...
            self.__value = something
            return

        @property
        def value(self) -> ValueType:
            if self.__value is UNSET:
                raise ValueError(
                    f"Please call `value_setter` before using this property."
                )
            return self.__value
    ```
    """

    __unset = None

    def __new__(cls) -> "Unset":
        """Return the singular instance of `Unset`."""
        if cls.__unset is None:
            cls.__unset = super().__new__(cls)
        return cls.__unset


UNSET = Unset()



"""This module contains the `Edge` and `Graph` classes."""

import typing

import numpy

from ..utils import constants
from ..utils import helpers
from . import cluster
from . import space

logger = helpers.make_logger(__name__)

AdjacencyDict = dict[cluster.Cluster, set["Edge"]]


class Edge:
    """An `Edge` represents a relationship between two `Cluster`s in a `Graph`.

    Two clusters in a graph have an edge between them if there is any overlap
    between the volumes of the two clusters.

    In CLAM, we do not allow edges from a cluster to itself.
    In CLAM, edges are bidirectional.
    """

    __slots__ = [
        "__left",
        "__right",
        "__distance",
    ]

    def __init__(
        self,
        left: cluster.Cluster,
        right: cluster.Cluster,
        distance: float,
    ) -> None:
        """Create an edge using two clusters and the distance between their centers."""
        self.__left, self.__right = (left, right) if left < right else (right, left)
        self.__distance = distance

    def __eq__(self, other: "Edge") -> bool:  # type: ignore[override]
        """Two edges are equal if their clusters are equal."""
        return self.clusters == other.clusters

    def __str__(self) -> str:
        """Returns a string representation of the edge."""
        return f"{self.__left!s} -- {self.__right!s}"

    def __repr__(self) -> str:
        """Returns a string representation of the edge."""
        return f"{self.__left!s} -- {self.__right!s} -- {self.__distance:.6e}"

    def __hash__(self) -> int:
        """Returns a hash of the edge."""
        return hash(str(self))

    def __contains__(self, c: cluster.Cluster) -> bool:
        """Returns whether the given cluster is in the edge."""
        return (c == self.__left) or (c == self.__right)

    @property
    def clusters(self) -> tuple[cluster.Cluster, cluster.Cluster]:
        """Returns the two clusters in the edge."""
        return self.__left, self.__right

    @property
    def left(self) -> cluster.Cluster:
        """Returns the left cluster in the edge."""
        return self.__left

    @property
    def right(self) -> cluster.Cluster:
        """Returns the right cluster in the edge."""
        return self.__right

    @property
    def distance(self) -> float:
        """Returns the distance between the centers of the two clusters."""
        return self.__distance

    @property
    def to_self(self) -> bool:
        """Whether the `left` and `right` clusters are equal."""
        return self.left == self.right

    def neighbor(self, c: cluster.Cluster) -> cluster.Cluster:
        """Returns the neighbor of the given cluster, if the cluster is in the edge.

        Otherwise raises a ValueError.
        """
        if c == self.__right:
            return self.__left

        if c == self.__left:
            return self.__right

        msg = f"Cluster {c.name} is not in this edge {self!s}."
        raise ValueError(msg)


class Graph:
    """A `Graph` of `Cluster`s with `Edge`s between clusters with overlapping volumes.

    We can select clusters from a tree to build a graph.

    All clusters in a graph must come from the same tree and the same
    `MetricSpace`. If using any of the `SelectionCriteria` from CLAM, this
    property will hold. Otherwise, it is upon the user to ensure.

    In CLAM, a `Graph` has a useful invariant:
        - Every instance in the dataset is in exactly one cluster in the graph.
        - No cluster is an ancestor/descendent of another cluster in the graph.

    If clusters were selected using any of the `SelectionCriteria`, this
    invariant will hold. Otherwise, the user must verify this manually.
    This invariant does not hold for sub-graphs such as connected components.
    """

    __slots__ = [
        "__clusters",
        "__root",
        "__metric_space",
        "__population",
        "__min_depth",
        "__max_depth",
        "__edges",
        "__adjacency_dict",
        "__neighborhood_sizes",
        "__components",
        "__distance_matrix",
    ]

    def __init__(
        self,
        clusters: set[cluster.Cluster],
        edges: typing.Optional[set[Edge]] = None,
    ) -> None:
        """Creates a graph from the given clusters.

        The edges need not be provided because we have efficient internal algorithms
        for finding all edges.

        Args:
            clusters: The set of clusters with which to build the graph.
            edges: Optional. The set of edges to use.
        """
        if len(clusters) == 0:
            msg = "Cannot create a graph with an empty set of clusters."
            raise ValueError(msg)

        self.__clusters = clusters

        _c = next(iter(clusters))
        self.__root = _c if _c.depth == 0 else _c.ancestry[0]
        self.__metric_space = _c.metric_space

        self.__population = sum(c.cardinality for c in self.__clusters)
        self.__min_depth = min(c.depth for c in self.__clusters)
        self.__max_depth = max(c.depth for c in self.__clusters)

        self.__edges: typing.Union[set[Edge], constants.Unset] = (
            edges or constants.UNSET
        )
        self.__adjacency_dict: typing.Union[
            AdjacencyDict,
            constants.Unset,
        ] = constants.UNSET
        self.__neighborhood_sizes: dict[cluster.Cluster, list[int]] = {}
        self.__components: typing.Union[
            list["Graph"],
            constants.Unset,
        ] = constants.UNSET
        self.__distance_matrix: typing.Union[
            numpy.ndarray,
            constants.Unset,
        ] = constants.UNSET

    @property
    def root(self) -> cluster.Cluster:
        """The root of the tree from which the clusters in the graph were selected."""
        return self.__root

    @property
    def clusters(self) -> set[cluster.Cluster]:
        """The set of clusters in the graph."""
        return self.__clusters

    @property
    def edges(self) -> set[Edge]:
        """The set of edges in the graph."""
        if self.__edges is constants.UNSET:
            msg = "Please call `build` before accessing this property."
            raise ValueError(msg)
        return self.__edges  # type: ignore[return-value]

    @property
    def metric_space(self) -> space.Space:
        """The metric space in which the clusters in the graph were selected."""
        return self.__metric_space

    @property
    def vertex_cardinality(self) -> int:
        """The number of clusters in the graph."""
        return len(self.__clusters)

    @property
    def edge_cardinality(self) -> int:
        """The number of edges in the graph."""
        return len(self.edges)

    @property
    def population(self) -> int:
        """The total number of instances across all clusters in the graph."""
        return self.__population

    @property
    def min_depth(self) -> int:
        """The minimum depth of any cluster in the graph."""
        return self.__min_depth

    @property
    def max_depth(self) -> int:
        """The maximum depth of any cluster in the graph."""
        return self.__max_depth

    @property
    def depth_range(self) -> tuple[int, int]:
        """The range of depths of the clusters in the graph."""
        return self.__min_depth, self.__max_depth

    @property
    def adjacency_dict(self) -> AdjacencyDict:
        """A dictionary of clusters and their sets of edges."""
        if self.__adjacency_dict is constants.UNSET:
            msg = "Please call `build` before accessing this property."
            raise ValueError(msg)
        return self.__adjacency_dict  # type: ignore[return-value]

    @property
    def indices(self) -> list[int]:
        """Indices of instances in all clusters in the graph."""
        return [i for c in self.__clusters for i in c.indices]

    @property
    def diameter(self) -> int:
        """The maximum eccentricity of any cluster in the graph."""
        return max(self.eccentricity(c) for c in self.__clusters)

    @property
    def components(self) -> list["Graph"]:
        """List of connected components, as sub-`Graph`s of this graph.

        If there is more than one component, no component will uphold the graph
        invariant.
        """
        if self.__components is constants.UNSET:
            self.__components = []
            unvisited = self.__clusters.copy()

            while len(unvisited) > 0:
                start = unvisited.pop()
                visited, _ = self.__traverse(start)
                unvisited = unvisited - visited
                self.__components.append(self.subgraph_with(visited))

        return self.__components  # type: ignore[return-value]

    @property
    def distance_matrix(self) -> tuple[list[cluster.Cluster], numpy.ndarray]:
        """A list of clusters in the graph and a square matrix of edge lengths."""
        if self.__distance_matrix is constants.UNSET:
            clusters = list(self.__clusters)
            indices: dict[cluster.Cluster, int] = {c: i for i, c in enumerate(clusters)}

            matrix = numpy.zeros(
                shape=(self.vertex_cardinality, self.vertex_cardinality),
            )
            for e in self.__edges:  # type: ignore[union-attr]
                i, j = indices[e.left], indices[e.right]
                matrix[i, j] = e.distance
                matrix[j, i] = e.distance

            self.__distance_matrix = matrix.astype(numpy.float32)

        return list(self.__clusters), self.__distance_matrix

    @property
    def adjacency_matrix(self) -> tuple[list[cluster.Cluster], numpy.ndarray]:
        """A list of clusters in the graph and a binary square matrix of edges."""
        clusters, matrix = self.distance_matrix
        matrix = (matrix > 0).astype(numpy.uint8)
        return clusters, matrix

    def build(self) -> "Graph":
        """Builds the `edges` and `adjacency_dict` of the graph."""
        logger.info(
            f"Building graph with {self.vertex_cardinality} clusters in a depth "
            f"range of {self.depth_range} ...",
        )

        if self.__edges is constants.UNSET:
            self.__edges = {
                Edge(c, n, d)
                for c in self.__clusters
                for n, d in c.candidate_neighbors.items()
                if (n != c) and (n in self.__clusters) and (d <= (c.radius + n.radius))
            }

        self.__adjacency_dict = {c: set() for c in self.__clusters}
        for e in self.__edges:  # type: ignore[union-attr]
            self.__adjacency_dict[e.left].add(e)
            self.__adjacency_dict[e.right].add(e)

        return self

    def assert_contains(self, c: cluster.Cluster) -> None:
        """Raises a ValueError if `c` is not in the graph."""
        if c not in self.__clusters:
            msg = f"Cluster {c} is not in this Graph."
            raise AssertionError(msg)

    def jaccard(self, other: "Graph") -> float:
        """Compute the Jaccard similarity with another graph."""
        self_indices = set(self.indices)
        other_indices = set(other.indices)

        intersection = self_indices.intersection(other_indices)
        union = self_indices.union(other_indices)

        return len(intersection) / len(union)

    def vertex_degree(self, c: cluster.Cluster) -> int:
        """Returns the degree of the given cluster."""
        self.assert_contains(c)
        return len(self.adjacency_dict[c])

    def edges_of(self, c: cluster.Cluster) -> list[Edge]:
        """Returns the edges of the given cluster."""
        self.assert_contains(c)
        return list(self.adjacency_dict[c])

    def neighbors_of(self, c: cluster.Cluster) -> list[cluster.Cluster]:
        """Returns the neighbors of the given cluster."""
        self.assert_contains(c)
        return [e.neighbor(c) for e in self.adjacency_dict[c]]

    def edge_distances_of(self, c: cluster.Cluster) -> list[float]:
        """Returns the distances of the edges of the given cluster."""
        self.assert_contains(c)
        return [e.distance for e in self.adjacency_dict[c]]

    def __traverse(
        self,
        start: cluster.Cluster,
    ) -> tuple[set[cluster.Cluster], list[int]]:
        """Performs a traversal, in arbitrary order, starting at the given cluster.

        The traversal continues until no new clusters can be visited.

        Args:
            start: where the traversal will start.

        Returns:
            A 2-tuple of:
                - set of visited clusters.
                - list of number of clusters in the frontier at each iteration
                  of the traversal.
        """
        visited: set[cluster.Cluster] = set()
        frontier: set[cluster.Cluster] = {start}
        frontier_sizes = []

        while len(frontier) > 0:
            new_frontier = {
                n
                for c in frontier
                for n in self.neighbors_of(c)
                if not ((n in visited) or (n in frontier))
            }
            visited.update(frontier)
            frontier = new_frontier
            frontier_sizes.append(len(frontier))

        return visited, frontier_sizes

    def frontier_sizes(self, c: cluster.Cluster) -> list[int]:
        """Returns the number of clusters in each frontier during the traversal."""
        self.assert_contains(c)
        if c not in self.__neighborhood_sizes:
            _, n = self.__traverse(c)
            self.__neighborhood_sizes[c] = n
        return self.__neighborhood_sizes[c]

    def eccentricity(self, c: cluster.Cluster) -> int:
        """Returns the eccentricity of the given cluster."""
        return len(self.frontier_sizes(c))

    def subgraph_with(self, clusters: set[cluster.Cluster]) -> "Graph":
        """Returns a sub-graph with the given clusters."""
        for c in clusters:
            self.assert_contains(c)

        edges = {
            e
            for e in self.__edges  # type: ignore[union-attr]
            if (e.left in clusters) and (e.right in clusters)
        }
        return Graph(clusters, edges).build()

    def component_containing(self, c: cluster.Cluster) -> "Graph":
        """Returns the connected component containing the given cluster."""
        self.assert_contains(c)
        for component in self.components:
            if c in component.__clusters:
                return component

        msg = "Cluster is the graph but not in any component. This is a bug."
        raise ValueError(
            msg,
        )

    def as_dot_string(self) -> str:
        """Returns a dot string representation of the graph for graphviz."""
        raise NotImplementedError

    @staticmethod
    def from_dot_string(dot_string: str) -> "Graph":
        """Returns a graph from the given dot string."""
        raise NotImplementedError

    def __add(self, c: cluster.Cluster) -> None:
        """Adds the given cluster to the graph."""
        raise NotImplementedError

    def __remove(self, c: cluster.Cluster) -> None:
        """Removes the given cluster from the graph."""
        raise NotImplementedError

    def replace_clusters(
        self,
        removals: set[cluster.Cluster],
        additions: set[cluster.Cluster],
    ) -> "Graph":
        """Returns a new graph with the given clusters removed and added."""
        raise NotImplementedError


__all__ = [
    "AdjacencyDict",
    "Edge",
    "Graph",
]



"""This module defines the `Metric` class."""

import abc
import typing

import numpy
from scipy.spatial.distance import cdist

SCIPY_METRICS = typing.Literal[
    "braycurtis",
    "canberra",
    "chebyshev",
    "cityblock",
    "correlation",
    "cosine",
    "dice",
    "directed_hausdorff",
    "euclidean",
    "hamming",
    "jaccard",
    "jensenshannon",
    "kulsinski",
    "kulczynski1",
    "mahalanobis",
    "matching",
    "minkowski",
    "rogerstanimoto",
    "russellrao",
    "seuclidean",
    "sokalmichener",
    "sokalsneath",
    "sqeuclidean",
    "yule",
]


class Metric(abc.ABC):
    """This represents a general distance metric.

    Mathematically, this is a distance function, `f`, which obeys the following:
        - scalar-valued: `f(x, y)` is a Real number for all `x` and `y`
        - non-negativity: `f(x, y)` >= 0 for all `x` and `y`
        - identity: `f(x, y)` = 0 iff `x` = `y`
        - symmetry: `f(x, y)` = `f(y, x)`
        - triangle inequality (optional): `f(x, y)` <= `f(x, z)` + `f(z, y)`
          for all `x`, `y` and `z`.
    """

    def __init__(self, name: str) -> None:
        """Initialize the `Metric`."""
        self.name = name

    def __eq__(self, other: "Metric") -> bool:  # type: ignore[override]
        """Check if two metrics are identical."""
        return self.name == other.name

    def __str__(self) -> str:
        """Return a string representation of the metric."""
        return self.name

    def __repr__(self) -> str:
        """Return a string representation of the metric."""
        return self.name

    @abc.abstractmethod
    def one_to_one(self, left: typing.Any, right: typing.Any) -> float:  # noqa: ANN401
        """Compute the distance between `left` and `right`."""
        pass

    @abc.abstractmethod
    def one_to_many(
        self,
        left: typing.Any,  # noqa: ANN401
        right: typing.Any,  # noqa: ANN401
    ) -> numpy.ndarray:
        """Compute distances from `left` to instances in `right`."""
        pass

    @abc.abstractmethod
    def many_to_many(
        self,
        left: typing.Any,  # noqa: ANN401
        right: typing.Any,  # noqa: ANN401
    ) -> numpy.ndarray:
        """Compute distances from instances in `left` to those in `right`."""
        pass

    @abc.abstractmethod
    def pairwise(self, instances: typing.Any) -> numpy.ndarray:  # noqa: ANN401
        """Compute a 2d array of distances among each pair in `instances`."""
        pass


class ScipyMetric(Metric):
    """Wraps any scipy distance metric.

    See scipy.spatial.distance.cdist
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.distance.cdist.html.
    """

    def __init__(self, name: SCIPY_METRICS) -> None:
        """Initialize the `ScipyMetric`."""
        if name not in typing.get_args(SCIPY_METRICS):
            msg = f"`name` must be one of {SCIPY_METRICS}. Got {name} instead"
            raise ValueError(
                msg,
            )
        super().__init__(name)

    def one_to_one(
        self,
        left: numpy.ndarray,
        right: numpy.ndarray,
        **kwargs,  # noqa: ANN003
    ) -> float:
        """Compute the distance between `left` and `right`."""
        left = left[None, :]
        right = right[None, :]
        return float(self.many_to_many(left, right, **kwargs).squeeze())

    def one_to_many(
        self,
        left: numpy.ndarray,
        right: numpy.ndarray,
        **kwargs,  # noqa: ANN003
    ) -> numpy.ndarray:
        """Compute a 1d array of distances.

        `left` is a single instance and `right` is multiple instances.
        """
        left = left[None, :]
        return self.many_to_many(left, right, **kwargs)[0]

    def many_to_many(
        self,
        left: numpy.ndarray,
        right: numpy.ndarray,
        **kwargs,  # noqa: ANN003
    ) -> numpy.ndarray:
        """Compute a 2d array of distances.

        `left` and `right` are both multiple instances.
        """
        return cdist(left, right, metric=self.name, **kwargs)  # type: ignore

    def pairwise(
        self,
        instances: numpy.ndarray,
        **kwargs,  # noqa: ANN003
    ) -> numpy.ndarray:
        """Compute a 2d array of distances among each pair in `instances`."""
        return self.many_to_many(instances, instances, **kwargs)


__all__ = [
    "SCIPY_METRICS",
    "Metric",
    "ScipyMetric",
]
